/*
 * Midtomup
 * 
 * Convert Midi files to MUP input files
 */

#include <ctype.h>
#include <fcntl.h>
#include <errno.h>
#include <time.h>
#include "midifile.h"
#include "mtm.h"

#define TIE 1
#define NO_TIE 0
#define MEASURE -11
#define MAXVEL   500

typedef enum { UNSET, SCORE, MUSIC, STAFF } Mode;

// global variables - see extern definitions in mtm.h for comments
long bytecount;		// position in .midi file
unsigned char debugbuff[32];
Bstylep bstyles;
int accnotes[TABSIZE];
int sccnotes[TABSIZE][3];
Trackp *tracks;		/* main event tracks */
Tsigp tsigs;		/* head of list */
Tsigp tsigp;		/* tail or work point to tsig list */
Tsigp ctsig;		/* current signature */
Keyp keysigs;		/* head of list of key changes */
Keyp keysigp;		/* pointer into list of key changes */

int sccwhre;
int accwhre;
int triplet;
int velfudge;
int shortest;
int bwhre;
int maxtrk;		/* number of tracks defined in midifile 	*/
int ffactor = 0;	/* default quantization 			*/
int trackno = 0;	/* current track number				*/
int stretch = 0;	/* -s flag :: amount to be added to each length */
int debug = 0;		/* -d flag :: debug output enabled		*/
int zflag = 0;		/* -z flag :: discard very short notes 		*/
int lyricsFlag = 0;	/* -l flag :: output lyrics lines		*/
int warnFlag = 1;	/* default - issue warnings 			*/
int current_key = 999;	/* insist on setting key on changing to music mode */
int new_key = 0;	/* if nothing else, use C major... 		*/
int tsigForced = 0;	/* set if tsig is forced			*/
int tsigsIgnored = 0;	/* count of tsigs in midifile			*/
long ticks;		/* current scan position in piece 		*/
long bar_start;		/* start of current bar (in ticks) 		*/
long bar_end;		/* end of current bar (in ticks) 		*/
long piece_end;		/* end of piece (in ticks) 			*/
long discards = 0;	/* notes discarded (too short)			*/
Trackp ct;		/* current track 				*/
long startTime;	 	/* start of actual track, allowing for pickup 	*/
long division;		/* resolution, #ticks/beat 			*/
float pad = 0;		/* -p pad factor, e.g. 3.5 add 3.5 beats of
				   rest for a pickup bar 		*/
char ifile[80];		/* input midifile name				*/
FILE *out;		/* output file					*/
FILE *inf;		/* input midifile				*/

// local to this file
static int prevlen = -1;
static int cflag;	/* -c flag :=: suppress comment */
static int inclMidi= 0;	/* -m flag :=: include midi statement in output */
static Tlistp secondaries;	// list of text lines to be appended
static Mode modeval;		// current MUP mode
static int dfltkey=0;		// default key (if no keysigs found)
static int pmeasure = 1;	// bar numbering required
static int measure;		// bar number for output
static int donestaffs = 0;	// staffs has been produced by #header
static int conflicts;		// count of note unresolved conflicts
static int velct;
static long velt[MAXVEL];
static int velvel[MAXVEL];
static char pfile[80];		// parameter file name
static char ofile[80];		// output mup file name
static int tsigErrs = 0;	// set if tsig errors detected
static char *clefName[] = { "treble", "treble", "bass", "tenor", "alto" };

#ifndef VERSION
#define VERSION "Version unknown"
#endif
static char *version = "Version "VERSION;

extern char *optarg;
extern int optind;

void addTsig(long t, int n, int d) {
  Tsigp tsp, tspp = NULL;
  // check to see if we already have an entry at this time, and overwrite it if so.

  if (tsigForced) {
    tsigsIgnored++;
    return;		// ignore subsequent changes
  }
  tsp = tsigs;	
  while (tsp != NULL) {
    if (tsp->when == t) {
      tsp->num = n;
      tsp->dnom = d;
      return;
    }
    // if new entry should come before this, break here
    if (tsp->when > t) 
      break;
    tspp = tsp;
    tsp = tsp->next;
  }
  tsp = (Tsigp)malloc(sizeof(Tsig));
  tsp->when = t;
  tsp->num  = n;
  tsp->dnom = d;
  if (tsigs == NULL) {
    tsigs = tsp;
    tsp->next = NULL;
  }
  else {
    tsp->next = tspp->next;
    tspp->next = tsp;
  }
}

static void extractTsig(char *p) {
  int d=0, n=0;

  while (*p == ' ' || *p == '\t')
    p++;
  while (isdigit(*p))
    n = 10 * n + *p++ - '0';
  if (*p != '/')
    fprintf(stderr, "// invalid time signature\n");
  else
    p++;
  while (isdigit(*p))
    d = 10 * d + *p++ - '0';
  if (n > 0 && d > 0) {
    addTsig(0, n, d);
  }
}

static void forceTsig(char *p) {
  extractTsig(p);
  tsigForced = 1;
}

static void extractKeysig(char *p) {
  int n=0;
  p += 4;	// skip over 'key='
  while (isdigit(*p))
    n = 10 * n + *p++ - '0';
  if (*p == '&')
    n = -n;
  new_key = n;
}

static void doheader(FILE *fp) {
  char buf[80];
  char *p;

  while(fgets(buf, 80, fp) != '\0') {
    if (strncmp("#headend", buf, 8) == 0 ||
	strncmp("#endhead", buf, 8) == 0 )
      return;
    fprintf(out, "%s", buf);
    if (strncmp(buf, "score", 5) == 0)
      modeval = SCORE;
    else if (strncmp(buf, "music", 5) == 0)
      modeval = MUSIC;
    else if (strncmp(buf, "staff ", 6) == 0)
      modeval = STAFF;
    else {
      p = buf;
      while (*p == ' ' || *p == '\t')
	p++;
      if (strncmp(p, "staffs=", 6) == 0)
	donestaffs = 1;
      else if (strncmp(p, "time=", 5) == 0)
	extractTsig(p+5);			// get a default time sig
      else if (strncmp(p, "key=", 4) == 0)
	extractKeysig(p);		// get default new key sig
    }
  }
  fprintf(out, "// NO \"#headend\" STATEMENT?!\n");
}

void dosecondaries(FILE *fp) {
  Tlistp e, lp = NULL;
  char buf[80];

  while (fgets(buf, 80, fp) != NULL) {
    if (strncmp(buf, "#secend", 7) == 0 ||
	strncmp(buf, "#endsec", 7) == 0 )
      return;
    e = (Tlistp)malloc(sizeof(Tlist));
    e->text = strdup(buf);
    e->next = NULL;
    if (lp == NULL)
      secondaries = e;
    else
      lp->next = e;
    lp = e;
  }
}

#define SKIPBL while (*p == ' ' || *p == '\t') p++;

void dobeamstyle(FILE *fp) {
  int j, n, d;
  long staff;
  char buf[80], *p;
  Bstylep bsp, wp;

  while (fgets(buf, 80, fp) != '\0') {
    if (strncmp("#beamend", buf, 8) == 0 ||
	strncmp("#endbeam", buf, 8) == 0 )
      return;
    else {
      p = buf;
      SKIPBL;
      staff = 0;
      if (*p == '[') {	// staff list specified ??
	p++;
	SKIPBL;
	n = 0;
	while (isdigit(*p))
	  n = 10*n + *p++ - '0';
	staff = n;
	SKIPBL;
	if (*p != ']') {
	  fprintf(stderr, "mtm: bad staff list in #beamstyle directive\n");
	  continue;
	}
	p++;
	SKIPBL;
      }
      n=0;
      d=0;
      while (isdigit(*p))
	n = 10 * n + *p++ - '0';
      if (*p != '/') {
	fprintf(stderr, "mtm: bad time signature in #beamstyle directive\n");
	continue;
      }
      p++;
      while (*p == ' ' || *p == '\t')
	p++;
      while (isdigit(*p))
	d = 10 * d + *p++ - '0';
      if (*p != '=') {
	fprintf(stderr, "mtm: bad time signature in #beamstyle directive\n");
	continue;
      }
      p++;
      bsp = (Bstylep)malloc(sizeof(Bstyle));
      bsp->staff = staff;
      bsp->bnum = n;
      bsp->bdnom = d;
      j = 0;
      for (j=0; j<79 && *p!='\n'; j++)
	bsp->style[j] = *p++;
      bsp->style[j] = 0;
      if (staff == 0) {		// if not staff specific, put at front
	bsp->next = bstyles;
	bstyles = bsp;
      }
      else {			// if staff-specific, it needs to go at back
	bsp->next = NULL;
	wp = bstyles;
	while (wp != NULL && wp->next != NULL)
	  wp = wp->next;
	if (wp == NULL)
	  bstyles = bsp;	// no previous entries
	else
	  wp->next = bsp;	// append to tail of list
      }
      if (cflag == 0) {
	if (staff == 0)
	  fprintf(out, "// beamstyle %d/%d = %s\n", bsp->bnum,
		  bsp->bdnom, bsp->style);
	else {
	  fprintf(out, "// beamstyle [%d] %d/%d = %s\n",
		  bsp->staff, bsp->bnum, bsp->bdnom, bsp->style);
	}
      }
    }
  }
  fprintf(out, "// NO \"#beamend\" STATEMENT?!\n");
}

void dotranslate(char *buf) {
  int trk = 0, k;
  char *p, *q;
  Trackp ct;

  p = buf+10;	// point at start of parameters
  while (*p == ' ' || *p == '\t')
    p++;
  if (!isdigit(*p)) {
    fprintf(stderr, "bad syntax in #translate directive - ignored\n");
    return;
  }
  while (isdigit(*p))
    trk = trk * 10 + *p++ - '0';
  if (trk >= maxtrk) {
    fprintf(stderr, "track number too big in #translate directive - ignored\n");
    return;
  }
  while (*p == ' ' || *p == '\t')
    p++;
  if (*p != '=') {
    fprintf(stderr, "bad syntax in #translate directive - ignored\n");
    return;
  }
  p++;
  while (*p == ' ' || *p == '\t')
    p++;
  q = p + strlen(p) - 1;
  if (*q == '\n')
    q--;
  while (q > p && (*q == ' ' || *q == '\t'))
    q--;
  q[1] = 0;	// delete trailing whitespace
  k = strlen(p);
  if (k <= 0 || k > 62) {
    fprintf(stderr, "translation too long in #translate directive\n");
    return;
  }
  ct = tracks[trk];
  q = strchr(p, '/');		// are two maps provided
  if (q != NULL) {
    *q++ = 0;			// split into two, point q at second half
  }
  ct->map1 = strdup(p);
  if (q != NULL)
    ct->map2 = strdup(q);
  else
    ct->map2 = ct->map1;
  if (cflag == 0)
    fprintf(out, "// Translate track %d to %s.\n", trk, p);
  return;
}

void dokey(char *buf) {
  int k, trk, key;

  k = sscanf(buf, "#key %d=%d", &trk, &key);
  if (k != 2) {
    fprintf(stderr, "bad #key directive :%s\n", buf);
    return;
  }
  tracks[trk]->hkey = key;
  if (cflag == 0)
    fprintf(out, "// Key sig for track %d set to %d\n", trk, key);
  return;
}

static char *fmt_off(float v) {
  static char buff[16];
  char *p;
  
  v *= ctsig->dnom / 4.0;	// adjust for time signature
  sprintf(buff, "%f", v+1.0);
  p = buff + strlen(buff) - 1;
  while (*p == '0')
    p--;
  if (*p == '.')
    p--;
  p[1] = 0;
  return buff;
}

static void mode(Mode v) {
  if (modeval == v)
    return;
  switch (v) {
  case SCORE:
    fprintf(out, "score\n");
    break;
  case MUSIC:
    if (new_key != current_key) {
      if (modeval != SCORE)
	fprintf(out, "score\n");
      if (new_key < 0)
	fprintf(out, "\tkey=%d&\n", -new_key);
      else
	fprintf(out, "\tkey=%d#\n", new_key);
      current_key = new_key;
    }
    fprintf(out, "music\n");
    break;
  default:
    ;
  }
  modeval = v;
}

void usage(char *pgmname) {
  fprintf(stderr, "MTM version %s\n", VERSION);
  fprintf(stderr, "usage: %s [options] inputfile\n", pgmname);
  fprintf(stderr, "options currently available are:\n");
  fprintf(stderr, "\t-a quantum = specify quantisation factor\n");
  fprintf(stderr, "\t-c = inhibit comments in output\n");
  fprintf(stderr, "\t-d = produce debugging printouts\n");
  fprintf(stderr, "\t-h = print this usage info and exit\n");
  fprintf(stderr, "\t-k num = specify a default key in range -7..7\n");
  fprintf(stderr, "\t-m = include MIDI commands in output\n");
  fprintf(stderr, "\t-n = don't add measure number comments\n");
  fprintf(stderr, "\t-o filename = specify output file, '-' means stdout\n");
  fprintf(stderr, "\t-f filename = specify control file name\n");
  fprintf(stderr, "\t-p pad = pad first bar to handle lead-ins\n");
  fprintf(stderr, "\t-s stretch = add a fixed percentage to note length before\n");
  fprintf(stderr, "\t\t quantising. For example -s25 adds 25%% to each note.\n");
  fprintf(stderr, "\t-t timesig = force time signature, overriding any from the midifile.\n");
  fprintf(stderr, "\t-v = print version number and exit\n");
  fprintf(stderr, "\t-V velocity = substitute fixed velocity in output\n");
  fprintf(stderr, "\t-w = suppress warning messages\n");
  fprintf(stderr, "\t-z = delete very short notes\n");
  exit(1);
}

static void doclef(char *p) {
  int trk, clef;
  p += 5;
  while (*p == ' ' || *p == '\t')
    p++;
  trk = 0;
  if (!isdigit(*p)) {
    fprintf(stderr, "missing track number in #clef directive - ignored\n");
    return;
  }
  while (isdigit(*p))
    trk = 10*trk + *p++ - '0';
  while (*p == ' ' || *p == '\t')
    p++;
  if (*p != '=') {
    fprintf(stderr, "syntax error in #clef directive - ignored\n");
    return;
  }
  p++;
  while (*p == ' ' || *p == '\t')
    p++;
  if (strncmp(p, "treble", 6) == 0)
    clef = 1;
  else if (strncmp(p, "bass", 4) == 0)
    clef = 2;
  else if (strncmp(p, "tenor", 5) == 0)
    clef = 3;
  else if (strncmp(p, "alto", 4) == 0)
    clef = 4;
  else {
    fprintf(stderr, "syntax error in #clef directive - ignored\n");
    return;
  }
  if (trk < maxtrk) {
    tracks[trk]->clef = clef;
  }
  else
    fprintf(stderr, "track number(%d) in #clef directive is too big\n", trk);
}

static void inputParams(void) {
  // check to see if there is a parameter file,
  // possible candidate name in pfile, otherwise try .midtomup
  //
  FILE *fp;
  char buf[80], *p;

  fp = fopen(pfile, "r");
  if (fp == NULL) {
    strcpy(pfile, ".midtomup");
    fp = fopen(pfile, "r");
  }
  if (fp == NULL) {
    pfile[0] = 0;	// indicate no params read
  }
  else {
    fprintf(out, "// with parameters from %s\n", pfile);
    while (fgets(buf, 80, fp) != '\0') {
      p = buf;
      while (*p == ' ' || *p == '\t')
	p++;
      if (*p == '\n')		// ignore null lines
	continue;
      if (p[0] != '#' && p[0] != '/') {
	fprintf(stderr, "Parse error in %s: %s\n", ifile, buf);
	break;
      }
      if (p[0] == '/' && p[1] == '/') {
	fprintf(out, "%s", buf);
	continue;
      }
      if (strncmp(p, "#header", 7) == 0)
	doheader(fp);
      else if (strncmp(p, "#secondaries", 12) == 0)
	dosecondaries(fp);
      else if (strncmp(p, "#beamstyle", 10) == 0)
	dobeamstyle(fp);
      else if (strncmp(p, "#translate", 10) == 0)
	dotranslate(p);
      else if (strncmp(p, "#key", 4) == 0)
	dokey(p);
      else if (strncmp(p, "#clef", 5) == 0)
	doclef(p);
      else
	fprintf(stderr, "Unknown rc command in %s: %s\n", ifile, buf);
    }
    fclose(fp);
  }
}

void initTracks(void) {
  int i;
  Trackp t;
  Eventlistp elp;

  tracks = (Trackp *)calloc(maxtrk, sizeof(Trackp));
  t = (Trackp)calloc(maxtrk, sizeof(Track));
  elp = (Eventlistp)calloc(2*maxtrk, sizeof(Eventlist));

  for (i = 0; i < maxtrk; i++,t++) {
    t->notes = elp++;		// assign notes list
    t->mevnts = elp++;		// assign midi events list
    t->minp = 128;		// initialise min and max pitch
    t->maxp = -1;		// ready for testing
    t->num = i;			// useful to get back to track number
    tracks[i] = t;		// store pointer in vector
  }
}

static int qcomp(const void *a, const void *b) {
  Eventp ap = *(Eventp *)a;
  Eventp bp = *(Eventp *)b;
  if (ap->start < bp->start)
    return -1;
  else if (ap->start > bp->start)
    return 1;
  /* obvious cases sorted, what remains is a->start == b->start.
     dispose of PATCH and TEXT items first */
  if (ap->type != NOTE)
    return -1;
  if (bp->type != NOTE)
    return 1;
  /* both are notes, return shortest before longest */
  else if (ap->end < bp->end)
    return -1;
  else if (ap->end > bp->end)
    return 1;
  else
    /* both same length, return lowest pitch first */
    if (ap->pitch < bp->pitch)
      return -1;
  else if (ap->pitch > bp->pitch)
    return 1;
  else
    return 0;		/* don't care - two identical notes ???? */
}

static void sortList(Eventlistp l) {
  int j, len;
  Eventp *list, e;

  len = l->count;
  list = (Eventp *)malloc((len+1) * sizeof(Eventp));
  e = l->e;
  for (j=0; j<len; j++) {
    list[j] = e;
    e = e->next;
  }
  qsort(list, len, sizeof(Eventp), qcomp);
  l->e = list[0];
  for (j=0; j<len-1; j++) {
    list[j]->next = list[j+1];
    list[j+1]->prev = list[j];
  }
  list[j]->next = NULL;
  list[0]->prev = NULL;
  free(list);
  l->p = l->e;
}

static void sorttracks(Trackp *tracks) {
  int i;
  
  for (i=0; i<maxtrk; i++) {	// i scans through the given tracks
    if (tracks[i]->notes->count > 0) {
      sortList(tracks[i]->notes);
      sortList(tracks[i]->mevnts);
    }
  }
}

static char *maptrack(Trackp t, int flag) {
  /* see if a track mapping has been defined */
  static char buff[32];

  if (t->map1 != NULL) {
    if (flag == 0)
      return t->map1;
    else
      return t->map2;
  }
  sprintf(buff, "%d", t->num);
  return buff;
}

static void resetvels(void) {
  velct = 0;
}

static void output_vels(Trackp t) {
  int i;

  for (i=0; i<velct; i++) {
    if (velvel[i] != t->vel) {
      float offset = (float)(velt[i] - bar_start) / (float)division;

      if (velfudge == 0 && inclMidi) {
	fprintf(out, "\nmidi %s: %s \"onvelocity=%d\";",
		maptrack(t, 1), fmt_off(offset), velvel[i]);
	}
      t->vel = velvel[i];
    }
  }
}

typedef struct tab Table;
struct tab {
  int num;			// numerator
  int denom;			// divisor
  int trip;			// triplet flag
  int len;			// length in ticks
  char *name;			// name
};

static Table notelens[] = {
  { 4, 1, 0, 0, "1" },			// semibreve
  { 3, 1, 0, 0, "2." },			// dotted minim
  { 2, 1, 0, 0, "2"  },			// minim
  { 3, 2, 0, 0, "4." },			// dotted crochet
  { 1, 1, 0, 0, "4"  },			// crochet
  { 3, 4, 0, 0, "8." },			// dotted quaver
  { 1, 2, 0, 0, "8"  },			// quaver
  { 1, 3, 1, 0, "8"  },			// triplet quaver
  { 3, 8, 0, 0, "16."},			// dotted semiquaver
  { 1, 4, 0, 0, "16" },			// semiquaver
  { 1, 6, 1, 0, "16" },			// triplet semiquaver
  { 3,16, 0, 0, "32."},			// dotted 32
  { 1, 8, 0, 0, "32" },			// plain 32
  { 1,12, 1, 0, "32" },			// triplet 32
  { 3,32, 0, 0, "64."},			// dotted 64
  { 1,16, 0, 0, "64" },			// plain 64
  { 1,24, 1, 0, "64" } };		// triplet 64
#define TABLEN 17

static void initLenTab(void) {
  int i;
  Table *tp = notelens;

  for (i=0; i<TABLEN; i++,tp++) {
    tp->len = (division * tp->num) / tp->denom;
  }
}

static int plen(int len) {
  // add note length to output file (if different from previous note).
  // If the given length is greater than one of the recognised lengths
  // return the excess (to be added later as a tie).
  int i, diff, best=9999;
  Table *tp, *bp = NULL;
  
  if (len == prevlen) {
    if (triplet > 0)
      triplet++;
    return 0;		/* same as previous note(s) - omit */
  }
  tp = notelens;
  best = tp->len;
  bp = tp;
  for (i=0; i<TABLEN; i++,tp++) {
    diff = tp->len - len;
    if (diff >= 0 && diff < best) {
      best = diff;
      bp = tp;
    }
    if (tp->trip) {
      // check for double triplets
      diff = 2 * (tp->len - len);
      if (diff >= 0 && diff < best) {
	best = diff;
	bp = tp;
      }
    }
  }
  if (bp == NULL) {
    fprintf(stderr, "mtm: fatal internal error!\n");
    exit(1);
  }
  if (bp->trip) {
    fprintf(out, "{%s", bp->name);
    triplet++;
  }
  else
    fprintf(out, "%s", bp->name);
  prevlen = len;
  diff = len - bp->len;
  if (diff < 0)
    return 0;
  return diff;
}

static void dorest(int start, int trck, int len) {
  int i, off, min = shortest;
  int /*threes,*/ base, mod;

  if (len > 0) {
    /* start is # of ticks from bar start */
    /* threes = 0; */
    base = division * 4 / ctsig->dnom;	
    if (ctsig->num % 3 == 0) {		// if 3 a factor? 
      /* threes = 1; */
      base *= 3;
    }
    while (len > base) {
      mod = start % base;
      while (mod != 0) {
	dorest(start, trck, base-mod);
	start += (base-mod);
	len -= (base-mod);
	mod = start % base;
      }
      off = base;
      if (len > base) {
	while (off * 2 < len && (start % (off*2)) == 0 )
	  off *= 2;
	plen(off);
	if (bar_start == 0 && start < startTime)
	  fprintf(out, "s;");
	else
	  fprintf(out, "r;");
	start += off;
	len -= off;
      }
    }
    while (len > 0) {
      i = start / min;
      off = min;
      if (i == 0) {			/* any len OK */
	while (off * 2 <= len)	/* double while less than length */
	  off *= 2;
      }	
      else
	while ((i & 1) == 0 && (off * 2) <= len) {
	  i = i >> 1;
	  off *= 2;
	}
      /* off is now largest factor of start, or == len */
      if (off != 0) {
	plen(off);
	len -= off;
	if (bar_start == 0 && start < startTime)
	  fprintf(out, "s;");
	else
	  fprintf(out, "r;");
	start += off;
      }
    }
  }
  else if (bar_start == 0 && start < startTime) {
    dorest(start, trck, startTime-start);
    dorest(startTime, trck, bar_end-startTime);
  }
  else
    fprintf(out, "mr;");
}

static void fixvel(long time, int vel) {
  int i;
  for (i=0; i<velct; i++) {
    if (velt[i] == time) {
      /* multiple notes at same time */
      if (velvel[i] < vel)
	velvel[i] = vel;
      return;
    }
  }
  if (velct >= MAXVEL) {
    fprintf(stderr, "mtm: velocity table overflow\n");
    exit(1);
  }
  if (velct == 0 || velvel[velct-1] != vel) {
    velt[velct] = time;
    velvel[velct++] = vel;
  }
}


static void putOctave(int nval) {
  switch (nval) {
  case 0:
    fprintf(out, "-----");
    break;
  case 1:
    fprintf(out, "----");
    break;
  case 2:
    fprintf(out, "---");
    break;
  case 3:
    fprintf(out, "--");
    break;
  case 4:
    fprintf(out, "-");
    break;
  case 5:
    break;
  case 6:
    fprintf(out, "+");
    break;
  case 7:
    fprintf(out, "++");
    break;
  case 8:
    fprintf(out, "+++");
    break;
  case 9:
    fprintf(out, "++++");
    break;
  case 10:
    fprintf(out, "+++++");
    break;
  }
}

void poctave(Trackp t, int nval) {
  /* for defoct=4 */
  if (t->key < -5)
    nval++;
  else if (t->key > 6)
    nval--;
  nval /= 12;
  if (t->clef == 2)
    nval++;
  putOctave(nval);
}

int goctave(Trackp t, int nval) {
  /* for defoct=4 */
  if (t->key < -5)
    nval++;
  else if (t->key > 6)
    nval--;
  nval /= 12;
  if (t->clef == 2)
    nval++;	// handle defoct=3
  return nval;
}

static void donote(Trackp t, Eventp np, int num, int len) {
  // output a chord, with num notes, length=len, np->first event
  //  remaining events should be in sequence following...
  Eventp e;
  int i;
  char pitches[128];	// pitches used in this chord
  if (len == 0)
    return;		// ignore zero-length notes!
  do {
    int left = plen(len);

    len = len - left;
    memset(pitches, 0, 128);	// set no notes set
    for (e=np, i=0; i<num; i++, e=e->next) {
      if (e->remains >= e->end) {
	e->remains += len;
	continue;
      }
      fixvel(e->remains, e->vel);
      e->remains += len;
      if (pitches[e->pitch] == 0) {
	pitches[e->pitch] = 1;
	putNote(t, e);
	//	e->flags |= F_TIE;	// set flag to note tie
	accwhre++;
	if (accwhre >= TABSIZE) {
	  fprintf(stderr, "oops!\n");
	  exit(1);
	}
	accnotes[accwhre] = e->pitch;
	putOctave(e->oct);
	//	if (left != 0 || e->remains < e->end ) {
	if (e->flags & F_TIEF || e->remains < e->end ) {
	  fprintf(out, "~");
	  if (triplet == 3) {
	    fprintf(out, "}3");
	    triplet = 0;
	  }
	  t->tied = 1;
	}
      }
    }
    fprintf(out, ";");
    len = left;
  } while (len != 0);
  if (triplet == 3) {
    fprintf(out, "}3;");
    triplet = 0;
  }
}

static void do_notes(int i, Trackp ct) {
  // output notes for the given track in the current bar
  int j;
  Eventp np = ct->notes->p;

  resetvels();
  if (np != NULL && np->type != END) {
    for (j = 0; j < maxtrk; j++) {
      accnotes[j] = -1;
      sccnotes[j][0] = -1;
      sccnotes[j][1] = -1;
      sccnotes[j][2] = -1;
    }
    accwhre = -1;
    sccwhre = -1;
    ticks = bar_start;
    /*
     * Do the requested track translations.
     */
    fprintf(out, "\n%s:", maptrack(tracks[i], 0));
    /* Is there a note in the measure? */
    if (np->remains >= ticks && np->remains < bar_end) {
      do {
	if (np == NULL || np->type == END) {	/* end of list */
	  /* no more notes left, see if padding needed */
	  if (ticks < bar_end) {
	    dorest(ticks-bar_start, i, bar_end - ticks);
	    ticks = bar_end;
	  }
	}
	else if (np->type == NOTE) {
	  /* No, does a note start on current tick? */
	  if (np->remains != ticks) {
	    /* Nope, so output a rest */
	    if (np->remains >= bar_end) {
	      dorest(ticks-bar_start, i, bar_end-ticks);
	      ticks = bar_end;
	      break;
	    }
	    else {
	      dorest(ticks-bar_start, i, np->remains-ticks);
	      ticks = np->remains;
	    }
	  }
	  /* having "rested" up to start of note... */
	  if (ticks < bar_end) {
	    Eventp nxp = np->next;
	    int num = 1;
	    long fin /*, len*/ ;

	    fin = np->end;	// end of first note here
	    // see how many notes start at same time
	    while (nxp != NULL && np->remains == nxp->remains) {
	      if (nxp->end > ticks && nxp->end < fin)
		fin = nxp->end;
	      nxp = nxp->next;
	      num++;
	    }
	      /* fin now earliest finish time */
	    if (nxp != NULL && nxp->remains < fin)
	      fin = nxp->remains;
	    /* Note is tied to a note not in this measure */
	    if (fin > bar_end)
	      fin = bar_end;
	    // len = np->end - np->remains;	// length remaining
	    donote(ct, np, num, fin-ticks);	// output 'num' notes
	    ticks = fin;
	    // np = nxp;
	  }
	  /* Next note in track, skip any zero of negative-length */
	  while (np->type == NOTE && np->remains >= np->end)
	    np = np->next;
	}
	else {	/* PATCH or TEXT */
	  if (np->type != END)
	    np = np->next;
	}
      } while (ticks < bar_end && np->remains < bar_end && np->type != END);
      if (ticks < bar_end) {
	dorest(ticks-bar_start, i, bar_end - ticks);
	ticks = bar_end;
      }
    }
    else {
      dorest(0, i, MEASURE);		// do measure rest
    }
  }
  ct->notes->p = np;
  output_vels(ct);
}

void addKey(long t, int k) {
  Keyp kp = (Keyp)malloc(sizeof(Key));
  kp->next = NULL;
  kp->when = t;
  kp->nkey = k;
  if (keysigs == NULL)
    keysigs = kp;
  else
    keysigp->next = kp;
  keysigp = kp;
}

static void do_args(int argc, char **argv) {
  char *outfname=NULL, *p;
  char *errmess;
  int c;
  char *notelistfile = NULL;
  static char nlf[256];
  FILE *tmp;

  pfile[0] = 0;			// mark parameter file as unspecified
  while ((c = getopt(argc, argv, "a:cdf:hk:l:mnN:o:p:s:t:vV:wz")) != EOF)
    switch(c) {
    case 'd':
      debug = 1;
      break;
    case 'k':
      dfltkey = atoi(optarg);
      if (dfltkey > 7 || dfltkey < -7) {
	fprintf(stderr, "mtm: default key setting out of range -7..7\n");
	exit(1);
      }
      addKey(0, dfltkey);
      break;
    case 'a':
      ffactor = atoi(optarg) / 4;
      break;
    case 'c':
      cflag = 1;		// inhibit most comments
      break;
    case 'f':			// specify parameter file name
      strcpy(pfile, optarg);
      break;
    case 'o':
      outfname = optarg;
      break;
    case 'p':
      pad = atof(optarg);
      break;
    case 's':
      stretch = atoi(optarg);
      break;
    case 'm':
      inclMidi = 1;
      break;
    case 'n':
      pmeasure = 0;
      break;
    case 'N':
      notelistfile = optarg;
      break;
    case  't':
      forceTsig(optarg);
      break;
    case 'v':
      fprintf(stderr, "%s\n", version);
      exit(0);
      break;
    case 'V':
      velfudge = atoi(optarg);
      break;
    case 'w':
      warnFlag = 0;
      break;
    case 'z':
      zflag = 1;
      break;
    case 'l':
      lyricsFlag = atoi(optarg);
      break;
    default:
      fprintf(stderr, "-%c not understood.\n", c);
    case 'h':
      usage(argv[0]);
      exit(1);
    }
  if (optind >= argc) {
    usage(argv[0]);
    exit(1);
  }
  inf = efopen(argv[optind]);
  if (pfile[0] == 0) {
    strcpy(pfile, ifile);
    p = strrchr(pfile, '.');
    if (p != NULL)
      *p = 0;
    strcat(pfile, ".mtm");
  }
  if (outfname == NULL) {
    strcpy(ofile, ifile);
    p = strrchr(ofile, '.');
    if (p != NULL)
      *p = 0;
    strcat(ofile, ".mup");
    outfname = ofile;
  }
  else 
    strcpy(ofile, outfname);
  if (strcmp(outfname, "-") == 0)
    out = stdout;
  else {
    out = fopen(outfname, "w");
    if (out == NULL) {
      errmess = strerror(errno);
      fprintf(stderr, "mtm: unable to open output file: %s\n", outfname);
      fprintf(stderr, "Reason: %s\n", errmess);
      exit(1);
    }
  }
  bytecount = -1;
  if (notelistfile == NULL) {
    snprintf(nlf, 256, "%s/.mtm/notelist", getenv("HOME"));
    tmp = fopen(nlf, "r");
    if (tmp == NULL)
#ifdef MTMLIB
      snprintf(nlf, 256, "%s/notelist", MTMLIB);
#else
      snprintf(nlf, 256, "/usr/local/lib/mtm/notelist");
#endif
    else
      fclose(tmp);
    notelistfile = nlf;
  }
  loadNoteList(notelistfile);
}

void compact(void) {
  // scan down the tracks from 1-31, discarding any with notect<=1
  // (note that all tracks have at least one event, type=END)
  // shunt non-zero tracks down to fill gaps, set trackno=new upper limit
  int i, k;
  for (i=1,k=1; i<trackno; i++) {
    if (tracks[i]->notes->count > 1) {
      if (i > k) {
	tracks[k]  = tracks[i];
	tracks[k]->num = k;		// update track number too
      }
      k++;
    }
  }
  trackno = k;
}
    
void dobar(void) {
  fprintf(out, "\nbar\n");
}

static void genMidi(Trackp t) {
  Eventp np = ct->mevnts->p;

  if (np != NULL && np->type != END) {
    /* first, scan bar for midi commands */
    while (np && np->start < bar_end) {
      float offset = (float)(np->start - bar_start) / (float)division;

      if (np->type == TEMPO)
	fprintf(out, "\nmidi all: %s \"tempo=%d\";",
		fmt_off(offset), np->pitch);
      else if (t != tracks[0]) {	// ignore other stuff on trk 0
	if (np->type == PATCH)
	  fprintf(out, "\nmidi %s: %s \"program=%d\";",
		  maptrack(t,1), fmt_off(offset), np->pitch);
	else if (np->type == TEXT )
	  fprintf(out, "\nmidi %s: %s \"text=%s\";",
		  maptrack(t,1), fmt_off(offset), (char *)(np->vel));
	else if (np->type == PITCH)
	  fprintf(out, "\nmidi %s: %s \"chanpressure=%d\";",
		  maptrack(t,1), fmt_off(offset), np->pitch);
	else if (np->type == PARAM)
	  fprintf(out, "\nmidi %s: %s \"param=%d,%d\";",
		  maptrack(t,1), fmt_off(offset), np->pitch, np->vel);
	else if (np->type == BEND)
	  fprintf(out, "\nmidi %s: %s \"text=bend=%d\";",
		  maptrack(t,1), fmt_off(offset), (np->pitch*128)+np->vel);
      }
      np = np->next;
    }
    ct->mevnts->p = np;
  }
}

static void formatDate(void) {
  // ctime does: Wed Nov 11 13:56:21 2005 (alter to suit local conditions)
  long tt = time(NULL);
  char *t = ctime(&tt);
  fputs(t, out);
}

static void doTimeSigs(int flag) {
  Bstylep bsp;

  while (tsigp != NULL && tsigp->when <= bar_start) {
    if (flag) {
      mode(SCORE);
      fprintf(out, "\ttime=%d/%d\n", tsigp->num, tsigp->dnom);
      for (bsp=bstyles; bsp!=NULL; bsp=bsp->next) {
	if (bsp->bnum == tsigp->num &&
	    bsp->bdnom == tsigp->dnom) {
	  if (bsp->staff == 0)
	    fprintf(out, "\tbeamstyle=%s\n", bsp->style);
	  else {
	    fprintf(out, "staff %d\n", bsp->staff);
	    fprintf(out, "\tbeamstyle=%s\n", bsp->style);
	    modeval = STAFF;
	  }
	}
      }
    }
    ctsig = tsigp;
    tsigp = tsigp->next;
  }
}

static void doKeySigs(int flag) {
  int i, j;
  Trackp t;

  while (keysigp != NULL && keysigp->when <= bar_start) {
    if (flag) {
      mode(SCORE);
      new_key = keysigp->nkey;
      //      if (keysigp->nkey < 0)
      //      	fprintf(out, "\tkey=%d&\n", -keysigp->nkey);
      //      else
      //      	fprintf(out, "\tkey=%d#\n", keysigp->nkey);
    }
    for (i = 1; i < trackno; i++) {
      t = tracks[i];
      j = keysigp->nkey + t->hkey;
      if (flag && t->notes->count > 0 && t->hkey != 0) {
	/*
	 * Print the staff number, remembering translations..
	 */
	fprintf(out, "staff %s\n", maptrack(t, 1));
	if (j <= 0)
	  fprintf(out, "\tkey=%d&\n", -j);
	else
	  fprintf(out, "\tkey=%d#\n", j);
      }
      t->key = j;
      setkeysig(t, j);
    }
    keysigp = keysigp->next;
  }
}
  
static int barlen(void) {
  // get length of current bar in ticks
  int ans = ctsig->num * division;

  switch (ctsig->dnom) {	// determine what the denominator is
  case 4:
    break;
  case 8:
    ans = ans / 2;
    break;
  case 2:
    ans *= 2;
    break;
  default:
    fprintf(stderr, "mtm: sorry, I can't handle this time signature\n");
    exit(1);
  }
  return ans;
}

static void scanNotes(Trackp t) {
  //	scan notes that start within this bar, 
  // 	converting pitch to octave,name,acc
  Eventp np = t->notes->p;
  int j;

  if (np == NULL || np->type == END)
    return;
  for (j = 0; j < maxtrk; j++) {
    accnotes[j] = -1;
    sccnotes[j][0] = -1;
    sccnotes[j][1] = -1;
    sccnotes[j][2] = -1;
  }
  accwhre = -1;
  sccwhre = -1;
  // any notes starting in this bar?
  while (np != NULL && np->remains < bar_end) {
    if (np->type == NOTE) {
      // only process actual notes
      getPitch(t, np);
    }
    np = np->next;
  }
  t->notes->p = np;	// save pointer for next bar
}

static void convertPitch(void) {
  int i;

  // setup first bar as current
  ticks = 0;
  bar_start = 0;

  /* Main loop to output notes */
  while (bar_start < piece_end) {
    doTimeSigs(0);		// if needed!
    doKeySigs(0);		// likewise
    bar_end = bar_start + barlen();
    // check for anomalous timesig changes in mid-bar
    if (tsigp != NULL && tsigp->when < bar_end) {
      bar_end = tsigp->when;	// force a short bar
    }
    /*
     * We will deal with one measure at a time, one track at a time.
     */
    for (i=0; i<trackno; i++)
      resetaccs(tracks[i]);		/* reset accidentals */
    for (i=0; i<trackno; i++)	{
      ct = tracks[i];
      prevlen = -1;			/* make sure we output length */
      /* reset to start of bar */
      if (i > 0)
	scanNotes(tracks[i]);
    }
    bar_start = bar_end;
  } 
}

static void modifyNote(Eventp ep) {
  // make an enharmonic change to the given note
  if (ep->acc == '#') {
    // note is x#, change to y& where y is next note up from x
    ep->note++;
    if (ep->note == 'h')
      ep->note = 'a';
    else if (ep->note == 'c')
      // we have moved across an octave boundary
      ep->oct++;
    ep->acc = '&';
  }
  else {
    // note is y&, change to x#
    if (ep->note == 'a')
      ep->note = 'g';
    else
      ep->note--;
    if (ep->note == 'b')
      ep->oct--;
    ep->acc = '#';
  }
  ep->flags |= F_MOD;
}

static void findClashes(void) {
  // we can do this a track at a time as we don't need to
  // track time sigs and key sigs etc.
  int i;
  Eventp ep, ep2;
  long end;
  int p0, p1;

  for (i=1; i<maxtrk; i++) {
    // scanning a particular track
    ct = tracks[i];
    ep = ct->notes->e;
    // scan events in this track
    while (ep != NULL) {
      if (ep->type == NOTE) {
	// we have a note, get end time and see if there are overlaps
	end = ep->end;				// record end of note time
	p0 = ep->oct * 7 + ep->note - 'a';	// and pitch (ignoring accs)
	ep2 = ep->next;				// get subsequent notes
	while (ep2 != NULL) {
	  if (ep2->type == END)
	    break;
	  if (ep2->type == NOTE) {
	    if (ep2->start >= end)
	      break;
	    p1 = ep2->oct * 7 + ep2->note - 'a';
	    if (p0 == p1) {
	      // we have a name clash - MUP would produce an error message
	      // if this was output. See if we can resolve the issue
	      if (ep->acc != ' ' && ep->acc != 'n') {
		// the first note is an accidental. We must modify it
		// if we are going to resolve the matter
		if ((ep->flags & F_MOD) != 0) {
		  // the note has already been modified to resolve another
		  // conflict. We cannot modify it back again
		  conflicts++;
		  goto NEXT;
		}
		modifyNote(ep);
	      }
	      else {
		// first note is natural, modify the second note if possible
		if ((ep2->flags & F_MOD) != 0)
		  conflicts++;
		else
		  modifyNote(ep2);
	      }
	    }
	  }
	  ep2 = ep2->next;
	}
      }
NEXT:
      ep = ep->next;
    }
  }
}

static void doLyrics(Eventp ep, int i, Trackp ct) {
  //	output lyrics for this track in this bar
  Eventp mep = ct->mevnts->p;
  Eventp lep = ct->notes->p;	// 1st note beyond this bar
  int done, any=0;
  
  while (ep != NULL && ep != lep) {
    if (ep->type == NOTE && (ep->flags & F_TIEB) == 0) {
      //    printf("Note: %ld\n", ep->start);
      while (mep != NULL && mep->start < ep->start) {
	mep = mep->next;
      }
      done = 0;
      if (any == 0) {
	if (ct->map1 == NULL)
	  fprintf(out, "\nL%d:	\"", i);
	else
	  fprintf(out, "\nL%s:	\"", ct->map1);
	any = 1;
      }
      while (mep != NULL && mep->start == ep->start) {
	if (mep->type == TEXT && mep->pitch == 5) {
	  fprintf(out, "%s ", (char *)mep->vel);
	  done = 1;
	}
	mep = mep->next;
      }
      if (done == 0)
	fprintf(out, "- ");
    }
    ep = ep->next;
  }
  if (any)
    fprintf(out, "\";\n");
  ct->mevnts->p = mep;
}

static void genMusic(void) {
  int i;
  Eventp ep;

  // setup first bar as current
  measure = 1;
  ticks = 0;
  bar_start = 0;
  initLenTab();

  /* Main loop to output notes */
  while (bar_start < piece_end) {
    doTimeSigs(1);		// if needed!
    doKeySigs(1);		// likewise
    mode(MUSIC);		// make sure we're in the right mode
    if (pmeasure)
      fprintf(out, "\n// bar %d", measure);
    measure++;
    bar_end = bar_start + barlen();
    // check for anomalous time sig changes in mid-bar
    if (tsigp != NULL && tsigp->when < bar_end) {
      fprintf(out, "// WHOOPS: time signature off bar-line\n");
      tsigErrs++;
      bar_end = tsigp->when;	// force a short bar
    }
    /*
     * We will deal with one measure at a time, one track at a time.
     */
    for (i=0; i<trackno; i++)
      resetaccs(tracks[i]);		/* reset accidentals */
    for (i=0; i<trackno; i++)	{
      triplet = 0;			/* ensure triplets start off */
      ct = tracks[i];
      prevlen = -1;			/* make sure we output length */
      if (inclMidi) {
	genMidi(ct);
      }
      /* reset to start of bar */
      if (i > 0) {			/* ignore stuff on the control track */
	ep = ct->notes->p;
	do_notes(i, ct);
	if (lyricsFlag & (1<<(i-1)))
	  doLyrics(ep, i, ct);
      }
    }
    dobar();
    bar_start = bar_end;
  }
}

static void doOutput(void) {
  int i;
  Tlistp tp;

  /*
   * First output a 'standard' Mup header message
   */
  fprintf(out, "//!Mup-Arkkra\n");
  fprintf(out, "// generated by mtm version %s on ", VERSION);
  formatDate();
  fprintf(out, "// from midi file %s\n", ifile);
  /*
   * Next, read the control file if present ...
   */
  inputParams();
  if (pfile[0] != 0)
    fprintf(out, "// end parameters from %s\n", pfile);
  sorttracks(tracks);
  compact();		// discard any tracks with notect=0 (apart from 0)
  
  /* Initialise track data */
  for (i=1; i<trackno; i++) {
    ct = tracks[i];
    ct->notes->p  = ct->notes->e;
    ct->mevnts->p = ct->mevnts->e;
    ct->tr = 0;
    ct->vel = -1;
    if (ct->clef == 0) {	// has it been defined in control file?
      if (ct->maxp < 65)
	ct->clef = 2;		// looks like bass clef to me
      else
	ct->clef = 1;		// safer to use Treble clef
    }
  }

  // set up time signature handling
  // If no time sig specified, insert one, set to 4/4
  if (tsigs == NULL) {
    addTsig(0, 4, 4);
  }
  ctsig = tsigs;		/* set first one as current */
  tsigp = tsigs;		/* reset pointer to start of list */

  // setup key signature handling
  keysigp = keysigs;		/* note number of key sigs read */
  //  if (keysigs == NULL) {	/* no key sigs in file? */
  for (i=0; i<trackno; i++)
    setkeysig(tracks[i], dfltkey);	/* set all tracks to C Major */
  //  }

  convertPitch();
  findClashes();

  /* re-Initialise track data */
  for (i=1; i<trackno; i++) {
    ct = tracks[i];
    ct->notes->p  = ct->notes->e;
    ct->mevnts->p = ct->mevnts->e;
    ct->tr  = 0;
    ct->vel = -1;
  }

  // set up time signature handling
  ctsig = tsigs;		/* set first one as current */
  tsigp = tsigs;		/* reset pointer to start of list */

  mode(SCORE);
  if (!donestaffs)
    fprintf(out, "\tstaffs=%d\n", trackno-1);	// add staffs directive
  tp = secondaries;
  while (tp) {
    fprintf(out, "%s", tp->text);
    tp = tp->next;
  }
  if (secondaries != NULL)
    modeval = UNSET;		/* secondaries may have left us in limbo? */
  for (i=1; i<trackno; i++) {
    if (tracks[i]->clef >= 2) {
      fprintf(out, "staff %s\n\tclef=%s\n",
	      maptrack(tracks[i], 1), clefName[tracks[i]->clef]);
      modeval = STAFF;
    }
    // otherwise assume treble
  }
  // setup key signature handling
  keysigp = keysigs;		/* note number of key sigs read */
  genMusic();			/* scan the piece, outputting the notes */
  if (discards > 0) {
    fprintf(stderr, "mtm: %ld note(s) discarded as too short\n", discards);
  }
  if (conflicts > 0) {
    fprintf(stderr, "mtm: %d note conflicts detected - MUP will complain\n",
	    conflicts);
  }
  if (warnFlag && tsigErrs > 0)
    fprintf(stderr, "mtm: %d time signature events off-bar detected\n",
	    tsigErrs);
  if (tsigsIgnored > 0) 
    fprintf(stderr, "mtm: %d time signature events ignored\n", tsigsIgnored);
}

int main(int argc, char **argv) {
  tsigs = NULL;
  bwhre = 0;
  stretch = 0;
  ffactor=16;		// reasonable default, can be overridden
  do_args(argc, argv);
  mfread();		// scan the midifile, building note events
  fclose(inf);
  /*
   * Done reading the midi file.  Now output the notes.
   */
  doOutput();
  exit(0);
}
